/*!
 * \file requester.cpp
 *
 * More elaborate description
 */
#include <ros/ros.h>
#include "koptplanner/inspection.h"
#include "shape_msgs/SolidPrimitive.h"
#include <cstdlib>
#include <visualization_msgs/Marker.h>
#include <fstream>
#include <geometry_msgs/PolygonStamped.h>
#include <nav_msgs/Path.h>
#include <std_msgs/Int32.h>
#include <ros/package.h>
#include "tf/tf.h"

std::vector<nav_msgs::Path> * readSTLfile(std::string name);

int main(int argc, char **argv)
{
  ros::init(argc, argv, "InspectionRequester");
  ROS_INFO("InspectionRequester is alive");
  if (argc != 1)
  {
    ROS_INFO("usage: plan");
    return 1;
  }

  ros::NodeHandle n;
  int _iterations;
  if(!n.getParam("InspectionRequester/iter", _iterations)){
    ROS_INFO("running with default number of iterations");
    _iterations = 20;
  }
  std::string mesh_name;
  if(!n.getParam("InspectionRequester/mesh", mesh_name)){
    ROS_INFO("Running with default mesh (not reccomended)");
    mesh_name = "fiveByFiveCube.stl";
  }
  ros::Publisher obstacle_pub = n.advertise<visualization_msgs::Marker>("scenario", 1);
  ros::Publisher stl_pub = n.advertise<nav_msgs::Path>("stl_mesh", 1);
  ros::ServiceClient client = n.serviceClient<koptplanner::inspection>("inspectionPath");

  ros::Rate r(50.0);
  ros::Rate r2(1.0);
  r2.sleep();

  /* define the bounding box */
  koptplanner::inspection srv;
	// Assuming a spherical workspace with diameter 1600 mm (reach of 1700 mm) maximal box constraints are  
  srv.request.spaceSize.push_back(0.9237*100 * 1.6); //in cm 
  srv.request.spaceSize.push_back(0.9237*100 * 1.6);
  //srv.request.spaceSize.push_back((0.9237/2) * 100); // Z > 0 not necessary.
	srv.request.spaceSize.push_back((0.9237/2 + 0.40) * 100 * 1.0); // results in area up to - 25 z.
  srv.request.spaceCenter.push_back(0);
  srv.request.spaceCenter.push_back(0);
	//srv.request.spaceCenter.push_back((0.9237/4) * 100);
  srv.request.spaceCenter.push_back(20); //center at 10 cm. 
  geometry_msgs::Pose reqPose;

  /* starting pose */
  reqPose.position.x = -30;
  reqPose.position.y = 30.0;
  reqPose.position.z = 20.0;
  tf::Quaternion q = tf::createQuaternionFromRPY(0.0, 1.4, 0.0);//1.4, 0.0);
  reqPose.orientation.x = q.x();
  reqPose.orientation.y = q.y();
  reqPose.orientation.z = q.z();
  reqPose.orientation.w = q.w();
  srv.request.requiredPoses.push_back(reqPose);

  /* final pose (remove if no explicit final pose is desired) */
  reqPose.position.x = -30;
  reqPose.position.y = 30.0;
  reqPose.position.z = 20.0;
  q = tf::createQuaternionFromRPY(0.0, 1.4,0.0);//1.4, 0.0);
  reqPose.orientation.x = q.x();
  reqPose.orientation.y = q.y();
  reqPose.orientation.z = q.z();
  reqPose.orientation.w = q.w();
  srv.request.requiredPoses.push_back(reqPose);

  /* parameters for the path calculation (such as may change during mission) */
  srv.request.incidenceAngle = M_PI/6.0;
  srv.request.minDist = 16;//16;
  srv.request.maxDist = 120;//120;
  ROS_INFO("selected %i iterations", _iterations);
  srv.request.numIterations = _iterations;

  /* read STL file and publish to rviz */
  std::vector<nav_msgs::Path> * mesh = readSTLfile(ros::package::getPath("request")+"/meshes/"+mesh_name);
  ROS_INFO("mesh size = %i", (int)mesh->size());
  for(std::vector<nav_msgs::Path>::iterator it = mesh->begin(); it != mesh->end() && ros::ok(); it++)
  {
    stl_pub.publish(*it);
    geometry_msgs::Polygon p;
    geometry_msgs::Point32 p32;
    p32.x = it->poses[0].pose.position.x;
    p32.y = it->poses[0].pose.position.y;
    p32.z = it->poses[0].pose.position.z;
    p.points.push_back(p32);
    p32.x = it->poses[1].pose.position.x;
    p32.y = it->poses[1].pose.position.y;
    p32.z = it->poses[1].pose.position.z;
    p.points.push_back(p32);
    p32.x = it->poses[2].pose.position.x;
    p32.y = it->poses[2].pose.position.y;
    p32.z = it->poses[2].pose.position.z;
    p.points.push_back(p32);
    srv.request.inspectionArea.push_back(p);
    r.sleep();
  }

  /* define obstacle regions as cuboids that are coordinate system aligned */
  
  shape_msgs::SolidPrimitive body;
  body.type = shape_msgs::SolidPrimitive::BOX;
  body.dimensions.push_back(40.0);
  body.dimensions.push_back(40.0);
  body.dimensions.push_back(130.0);
  srv.request.obstacles.push_back(body);
  geometry_msgs::Pose pose;
  pose.position.x = 0.0;
  pose.position.y = 0.0;
  pose.position.z = 0.0;
	
	// Kan vurdere å rotere 45 grader. 
  pose.orientation.x = 0.0;
  pose.orientation.y = 0.0;
  pose.orientation.z = 0.0;
  pose.orientation.w = 1.0;
  srv.request.obstaclesPoses.push_back(pose);
  srv.request.obstacleIntransparancy.push_back(0);

  // publish obstacles for rviz 
  visualization_msgs::Marker marker;
  marker.header.frame_id = "/kopt_frame";
  marker.header.stamp = ros::Time::now();
  marker.ns = "obstacles";
  marker.id = 0; // enumerate when adding more obstacles
  marker.type = visualization_msgs::Marker::CUBE;
  marker.action = visualization_msgs::Marker::ADD;

  marker.pose.position.x = pose.position.x;
  marker.pose.position.y = pose.position.y;
  marker.pose.position.z = pose.position.z;
  marker.pose.orientation.x = pose.orientation.x;
  marker.pose.orientation.y = pose.orientation.y;
  marker.pose.orientation.z = pose.orientation.z;
  marker.pose.orientation.w = pose.orientation.w;

  marker.scale.x = body.dimensions[0];
  marker.scale.y = body.dimensions[1];
  marker.scale.z = body.dimensions[2];

  marker.color.r = 0.0f;
  marker.color.g = 0.0f;
  marker.color.b = 1.0f;
  marker.color.a = 0.5;

  marker.lifetime = ros::Duration();
  obstacle_pub.publish(marker);
	
	r.sleep();
  
//Obstacle two.
  
	shape_msgs::SolidPrimitive wall;
  wall.type = shape_msgs::SolidPrimitive::BOX;
  wall.dimensions.push_back(5.0);
  wall.dimensions.push_back(180.0);
  wall.dimensions.push_back(180.0);
  srv.request.obstacles.push_back(body);
  //geometry_msgs::Pose pose;
  pose.position.x = 0.0;
  pose.position.y = -30.0;
  pose.position.z = 0.0;
	
	q = tf::createQuaternionFromRPY(0.0, 0.0,-M_PI/4.0);
	// Kan vurdere å rotere 45 grader. 
  pose.orientation.x = q.x();
  pose.orientation.y = q.y();
  pose.orientation.z = q.z();
  pose.orientation.w = q.w();
  srv.request.obstaclesPoses.push_back(pose);
  srv.request.obstacleIntransparancy.push_back(0);

  // publish obstacles for rviz 
  visualization_msgs::Marker marker2;
  marker2.header.frame_id = "/kopt_frame";
  marker2.header.stamp = ros::Time::now();
  marker2.ns = "obstacles";
  marker2.id = 0; // enumerate when adding more obstacles
  marker2.type = visualization_msgs::Marker::CUBE;
  marker2.action = visualization_msgs::Marker::ADD;

  marker2.pose.position.x = pose.position.x;
  marker2.pose.position.y = pose.position.y;
  marker2.pose.position.z = pose.position.z;
  marker2.pose.orientation.x = pose.orientation.x;
  marker2.pose.orientation.y = pose.orientation.y;
  marker2.pose.orientation.z = pose.orientation.z;
  marker2.pose.orientation.w = pose.orientation.w;

  marker2.scale.x = wall.dimensions[0];
  marker2.scale.y = wall.dimensions[1];
  marker2.scale.z = wall.dimensions[2];

  marker2.color.r = 0.0f;
  marker2.color.g = 0.0f;
  marker2.color.b = 1.0f;
  marker2.color.a = 0.5;

  marker2.lifetime = ros::Duration();
  obstacle_pub.publish(marker2);	

  r.sleep();
  

  if (client.call(srv))
  {
    /* writing results of scenario to m-file. use supplied script to visualize */
    std::fstream pathPublication;
    std::string pkgPath = ros::package::getPath("request");
    pathPublication.open((pkgPath+"/visualization/inspectionScenario.m").c_str(), std::ios::out);
    if(!pathPublication.is_open())
    {
      ROS_ERROR("Could not open 'inspectionScenario.m'! Inspection path is not written to file");
      return 1;
    }
    pathPublication << "inspectionPath = [";
    for(std::vector<geometry_msgs::PoseStamped>::iterator it = srv.response.inspectionPath.poses.begin(); it != srv.response.inspectionPath.poses.end(); it++)
    {
      tf::Pose pose;
      tf::poseMsgToTF(it->pose, pose);
      double yaw_angle = tf::getYaw(pose.getRotation());
      pathPublication << it->pose.position.x << ", " << it->pose.position.y << ", " << it->pose.position.z << ", 0, 0, " << yaw_angle << ";\n";
    }
    pathPublication << "];\n";
    pathPublication << "inspectionCost = " << srv.response.cost << ";\n";
    pathPublication << "numObstacles = " << std::min(srv.request.obstacles.size(),srv.request.obstaclesPoses.size()) << ";\n";
    for(int i = 0; i<std::min(srv.request.obstacles.size(),srv.request.obstaclesPoses.size()); i++)
    {
      pathPublication << "obstacle{" << i+1 << "} = [" << srv.request.obstacles[i].dimensions[0] << ", " << srv.request.obstacles[i].dimensions[1] << ", " << srv.request.obstacles[i].dimensions[2] << ";\n";
      pathPublication << srv.request.obstaclesPoses[i].position.x << ", " << srv.request.obstaclesPoses[i].position.y << ", " << srv.request.obstaclesPoses[i].position.z << "];\n";
    }
    pathPublication << "meshX = [";
    for(std::vector<nav_msgs::Path>::iterator it = mesh->begin(); it != mesh->end() && ros::ok(); it++)
    {
      pathPublication << it->poses[0].pose.position.x << ", " << it->poses[1].pose.position.x << ", " << it->poses[2].pose.position.x << ";\n";
    }
    pathPublication << "];\nmeshY = [";
    for(std::vector<nav_msgs::Path>::iterator it = mesh->begin(); it != mesh->end() && ros::ok(); it++)
    {
      pathPublication << it->poses[0].pose.position.y << ", " << it->poses[1].pose.position.y << ", " << it->poses[2].pose.position.y << ";\n";
    }
    pathPublication << "];\nmeshZ = [";
    for(std::vector<nav_msgs::Path>::iterator it = mesh->begin(); it != mesh->end() && ros::ok(); it++)
    {
      pathPublication << it->poses[0].pose.position.z << ", " << it->poses[1].pose.position.z << ", " << it->poses[2].pose.position.z << ";\n";
    }
    pathPublication << "];\n";
  }
  else
  {
    ROS_ERROR("Failed to call service planner");
    return 1;
  }

  return 0;
}

/**
*  \brief This function reads an ACII STL file for inspection planning
*/
std::vector<nav_msgs::Path> * readSTLfile(std::string name)
{
  std::vector<nav_msgs::Path> * mesh = new std::vector<nav_msgs::Path>;
  std::fstream f;
  f.open(name.c_str());
  assert(f.is_open());
  int MaxLine = 0;
  char* line;
  double maxX = -DBL_MAX;
  double maxY = -DBL_MAX;
  double maxZ = -DBL_MAX;
  double minX = DBL_MAX;
  double minY = DBL_MAX;
  double minZ = DBL_MAX;
  assert(line = (char *) malloc(MaxLine = 80));
  f.getline(line, MaxLine);
  if(0 != strcmp(strtok(line, " "), "solid"))
  {
    ROS_ERROR("Invalid mesh file! Make sure the file is given in ascii-format.");
    ros::shutdown();
  }
  assert(line = (char *) realloc(line, MaxLine));
  f.getline(line, MaxLine);
  int k = 0;
  while(0 != strcmp(strtok(line, " "), "endsolid") && !ros::isShuttingDown())
  {
    int q = 0;
    nav_msgs::Path p;
    geometry_msgs::PoseStamped v1;
    for(int i = 0; i<7; i++)
    {
      while(line[q] == ' ')
        q++;
      if(line[q] == 'v')
      {
        const double yawTrafo = 0.0;      // used to rotate the mesh before processing
        const double scaleFactor = 10;   // used to scale the mesh before processing (from mm to cm)
        const double offsetX = 35.0; //40.0       // used to offset the mesh before processing
        const double offsetY = -30.0; // -12       // used to offset the mesh before processing
        const double offsetZ = 30.0;       // used to offset the mesh before processing

        geometry_msgs::PoseStamped vert;
        char* v = strtok(line+q," ");
        v = strtok(NULL," ");
        double xtmp = atof(v)/scaleFactor;
        v = strtok(NULL," ");
        double ytmp = atof(v)/scaleFactor;
        vert.pose.position.x = cos(yawTrafo)*xtmp-sin(yawTrafo)*ytmp;
        vert.pose.position.y =  sin(yawTrafo)*xtmp+cos(yawTrafo)*ytmp;
        v = strtok(NULL," ");
        vert.pose.position.z =  atof(v)/scaleFactor;
        vert.pose.position.x -= offsetX;
        vert.pose.position.y -= offsetY;
        vert.pose.position.z -= offsetZ;
        if(maxX<vert.pose.position.x)
          maxX=vert.pose.position.x;
        if(maxY<vert.pose.position.y)
          maxY=vert.pose.position.y;
        if(maxZ<vert.pose.position.z)
          maxZ=vert.pose.position.z;
        if(minX>vert.pose.position.x)
          minX=vert.pose.position.x;
        if(minY>vert.pose.position.y)
          minY=vert.pose.position.y;
        if(minZ>vert.pose.position.z)
          minZ=vert.pose.position.z;
        vert.pose.orientation.x =  0.0;
        vert.pose.orientation.y =  0.0;
        vert.pose.orientation.z =  0.0;
        vert.pose.orientation.w =  1.0;
        p.poses.push_back(vert);
        if(p.poses.size() == 1)
          v1 = vert;
      }
      assert(line = (char *) realloc(line, MaxLine));
      f.getline(line, MaxLine);
    }
    p.poses.push_back(v1);
    p.header.frame_id = "/kopt_frame";
    p.header.stamp = ros::Time::now();
    p.header.seq = k;
    mesh->push_back(p);
    k++;
  }
  free(line);
  f.close();
  ROS_INFO("Mesh area is bounded by: [%2.2f,%2.2f]x[%2.2f,%2.2f]x[%2.2f,%2.2f]", minX,maxX,minY,maxY,minZ,maxZ);
  return mesh;
}
